import imp
from leafblade.utilities import CodeBlock, escape_content, escape_attribute_value, void_tags
from leafblade.parser import ElementNode, AttributeNode, StringNode, CodeNode

class Context:
	def __init__(self, attributes=None, content=None):
		self.attributes = attributes
		self.content = content

def compile_node(node, context):
	for special, callback in special_elements.items():
		if isinstance(node, special):
			return callback(node, context)

	if isinstance(node, ElementNode):
		if node.name in special_elements:
			return special_elements[node.name](node, context)

		new_context = Context(attributes=CodeBlock())
		is_void = node.name in void_tags # TODO: Require lowercase?

		if not is_void:
			new_context.content = CodeBlock()

		context.content.add_text("<" + node.name, None)

		for child in node.children:
			compile_node(child, new_context)

		context.content.add_part(new_context.attributes)
		context.content.add_text(">", None)

		if not is_void:
			context.content.add_part(new_context.content)
			context.content.add_text("</" + node.name + ">", None)
	elif isinstance(node, AttributeNode):
		context.attributes.add_text(" " + node.name, None)

		if node.value:
			context.attributes.add_text('="', None)
			context.attributes.add_part(node.value.content)
			context.attributes.add_text('"', None)
	elif isinstance(node, StringNode):
		context.content.add_part(node.content)
	elif isinstance(node, CodeNode):
		context.content.add_code(node.code.lstrip())

		if node.children:
			context.content.indent()
			for child in node.children:
				compile_node(child, context)
			context.content.unindent()
	elif node.children:
		for child in node.children:
			compile_node(child, context)

def compile_tree(tree, **kwargs):
	code = CodeBlock()

	compiled_code = CodeBlock()
	compile_node(tree, Context(content=compiled_code))

	if all(part[0] == "text" for part in compiled_code.parts):
		code.add_code("static = " + repr("".join(part[1] for part in compiled_code.parts)))

		code.add_code("def write(__write, __data):")
		code.indent()
		code.add_code("__write(static)")
		code.unindent()

		code.add_code("def render(__data):")
		code.indent()
		code.add_code("return static")
		code.unindent()
	else:
		code.add_code("from leafblade.utilities import TemplateData, escape_content, escape_attribute_value")

		code.add_code("def write(__write, __data):")
		code.indent()
		code.add_code("data = TemplateData()")
		code.add_code("data.__dict__.update(__data)")
		code.add_part(compiled_code)
		code.unindent()

		code.add_code("def render(__data):")
		code.indent()
		code.add_code("output = []")
		code.add_code("write(output.append, __data)")
		code.add_code("return ''.join(output)")
		code.unindent()

	source = code.join()

	filename = kwargs.get("filename", "<string>")
	compiled = compile(source, filename, "exec")
	module = imp.new_module(filename)
	module.__name__ = filename
	exec(compiled, module.__dict__)

	return module

special_elements = {}

def special(t):
	def add_special(handler):
		special_elements[t] = handler
		return handler

	return add_special



from . import parser
from . import compiler

class DoctypeNode(parser.Node):
	pass

@parser.special("doctype")
def parse_doctype(self, c):
	self.context = DoctypeNode(self.current_indent, self.context)
	return self.content(c)

@compiler.special(DoctypeNode)
def compile_doctype(node, context):
	if node.children:
		raise SyntaxError("The doctype element cannot have children")

	context.content.add_text("<!DOCTYPE html>", None)



from leafblade.utilities import CodeBlock, escape_content, escape_attribute_value

class SyntaxError(SyntaxError):
	pass

class Node:
	def __init__(self, indent, parent):
		self.indent = indent
		self.parent = parent
		self.children = []

		if parent:
			parent.add_child(self)

	def add_child(self, child):
		self.children.append(child)

class RootNode(Node):
	pass

class NamedNode(Node):
	def __init__(self, indent, parent, name):
		super().__init__(indent, parent)
		self.name = name

class ElementNode(NamedNode):
	pass

class AttributeNode(NamedNode):
	value = None

	def __init__(self, indent, parent, name):
		super().__init__(indent, parent, name)
		self.children = None

	def add_child(self, child):
		raise SyntaxError("Attributes cannot have children")

class StringNode(Node):
	def __init__(self, indent, parent):
		super().__init__(indent, parent)
		self.content = CodeBlock()
		self.children = None

	def add_child(self, child):
		raise SyntaxError("Strings cannot have children")

class CodeNode(Node):
	code = ""

class Parser:
	line = 1
	offset = 0
	current_line = ""
	current_indent = 0

	def __init__(self, **kwargs):
		self.filename = kwargs.get("filename", "<string>")
		self.tree = RootNode(-1, None)
		self.context = self.tree

	def content(self, c):
		if c == "\n":
			self.current_indent = 0
			return self.indent
		if c == "#":
			return self.comment
		if c == "%":
			self.context = CodeNode(self.current_indent, self.context)
			return self.code
		if c == "!":
			if isinstance(self.context, AttributeNode):
				raise SyntaxError("Attributes cannot have raw strings as values")

			return self.raw_string
		if c == '"':
			string = StringNode(None, None)
			string.parent = self.context

			if isinstance(self.context, AttributeNode):
				self.context.value = string
				self.string_escape = escape_attribute_value
			else:
				self.context.children.append(string)
				self.string_escape = escape_content

			self.context = string

			return self.string
		if c.isspace():
			return self.content
		if c.isalpha():
			self.current_identifier = ""
			return self.identifier(c)

		raise SyntaxError("Unexpected {!r}".format(c))

	def comment(self, c):
		if c == "\n":
			return self.content(c)

		return self.comment

	def code(self, c):
		if c == "\n":
			return self.content(c)

		self.context.code += c
		return self.code

	def unindent(self):
		while self.context.indent is None or self.current_indent <= self.context.indent:
			self.context = self.context.parent

		if self.current_indent > self.context.indent + 1:
			raise SyntaxError("Excessive indent")

	def indent(self, c):
		if c == "\n":
			# TODO: Print trailing whitespace warning if current_indent > 0?
			self.current_indent = 0
			return self.indent

		if c != "\t":
			self.unindent()
			return self.content(c)

		self.current_indent += 1
		return self.indent

	def identifier(self, c):
		if c == ":":
			return self.colon
		if c == "-" or c.isalnum():
			self.current_identifier += c
			return self.identifier

		special = special_elements.get(self.current_identifier)
		if special:
			return special(self, c)

		self.context = ElementNode(self.current_indent, self.context, self.current_identifier)
		return self.content(c)

	def colon(self, c):
		if c.isalpha():
			self.current_identifier += ":" + c
			return self.identifier(c)

		self.context = AttributeNode(None, self.context, self.current_identifier)
		return self.content(c)

	def string(self, c):
		if c == '"':
			self.unindent()
			return self.content
		if c == "\\":
			return self.escape
		if c == "#":
			return self.enter_interpolation

		self.context.content.add_text(c, self.string_escape)
		return self.string

	def raw_string(self, c):
		if c != '"':
			raise SyntaxError("Expected string")

		self.context = StringNode(None, self.context)
		self.string_escape = None

		return self.string

	def escape(self, c):
		if c != "'":
			self.context.content.add_text("\\", self.string_escape)

		self.context.content.add_text(c, self.string_escape)
		return self.string

	def enter_interpolation(self, c):
		if c == "{":
			self.current_expression = ""
			return self.interpolation

		self.context.content.add_text("#" + c, self.string_escape)
		return self.string

	def interpolation(self, c):
		if c == "}":
			self.context.content.add_expression(self.current_expression, self.string_escape)
			return self.string

		self.current_expression += c
		return self.interpolation

	state = content

	def feed(self, text):
		try:
			for i, c in enumerate(text):
				if c == "\n":
					self.line += 1
					self.offset = 0
					self.current_line = text[i+1:].partition("\n")[0]
				else:
					self.offset += 1

				self.state = self.state(c)

			if c != "\n":
				self.state("\n")
		except SyntaxError as error:
			error.filename = self.filename
			error.lineno = self.line
			error.offset = self.offset
			error.text = self.current_line
			raise

def parse(template, **kwargs):
	parser = Parser(**kwargs)
	parser.feed(template)
	return parser.tree

special_elements = {}

def special(name):
	def add_special(handler):
		special_elements[name] = handler
		return handler

	return add_special



void_tags = {
	"area", "base", "br", "col", "command", "embed", "hr", "img", "input",
	"keygen", "link", "meta", "param", "source", "track", "wbr"
}

def escape_literal(text):
	return text.replace("\n", r"\n").replace("\r", r"\r").replace("'", r"\'")

try:
	from leafblade_escape import escape_content, escape_attribute_value
except ImportError as e:
	def escape_content(text):
		return text.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")

	def escape_attribute_value(text):
		return text.replace("&", "&amp;").replace('"', "&quot;")

class CodeBlock:
	def __init__(self):
		self.parts = []

	def add_text(self, text, escape_function):
		self.parts.append(("text", text, escape_function))

	def add_expression(self, expression, escape_function):
		self.parts.append(("expression", expression, escape_function))

	def add_code(self, code):
		self.parts.append(("code", code, None))

	def add_part(self, part):
		self.parts.extend(part.parts)

	def indent(self):
		self.parts.append(("indent", None, None))

	def unindent(self):
		self.parts.append(("unindent", None, None))

	def join(self, state="code"):
		code = ""
		indent_level = 0

		def indent():
			return "\t" * indent_level

		for t, part, escape_function in self.parts:
			if t == "text":
				if state == "expression":
					code += ")\n" + indent() + "__write('"
				elif state == "code":
					code += indent() + "__write('"

				if escape_function:
					code += escape_literal(escape_function(part))
				else:
					code += escape_literal(part)
			elif t == "expression":
				if state == "text":
					code += "')\n" + indent() + "__write("
				elif state == "code":
					code += indent() + "__write("

				if escape_function:
					code += escape_function.__name__ + "(str((" + part + ")))"
				else:
					code += "str((" + part + "))"
			elif t == "code":
				if state == "text":
					code += "')\n"
				elif state == "expression":
					code += ")\n"

				code += indent() + part + "\n"
			else:
				if state == "text":
					code += "')\n"
				elif state == "expression":
					code += ")\n"

				if t == "indent":
					indent_level += 1
				elif t == "unindent":
					indent_level -= 1
				else:
					raise ValueError("Unrecognized type {type!r}".format(type=t))

				state = "code"
				continue

			state = t

		if state == "text":
			code += "')\n"
		elif state == "expression":
			code += ")\n"

		return code

class TemplateData:
	pass



from distutils.core import setup, Extension

long_description="""
A Python port of `Razor Leaf`__.

__ https://github.com/charmander/razorleaf
"""

escape = Extension("leafblade_escape", sources=["leafblade/escape.c"])

setup(
	name="leaf-blade",
	version="0.1.0",
	packages=["leafblade"],
	ext_modules=[escape],
	description="A template engine for HTML",
	long_description=long_description,
	classifiers=[
		"Programming Language :: Python :: 3",
		"Topic :: Internet :: WWW/HTTP :: Dynamic Content",
		"Topic :: Text Processing :: Markup :: HTML",
		"Development Status :: 2 - Pre-Alpha"
	]
)
